/*
 * JESTER.C
 * Interface to search/move algorithms from SAN Kit
 * Note that files here are chess engine files, not SAN kit.
 *
 * 02-08-2001 17:07:25  Modify max_time in init() to be int
 *            17:27:00  Add print_board() from original
 * 02-09-2001  9:05:22  get_ms() returns long
 *            16:10:08  fix promotion piece in playmove()
 * 02-12-2001 10:46:18  rename sources and headers, stdtype.h
 *            13:30:37  initial gen() added to playmove()
 * 02-13-2001  9:16:15  add code to set up board
 *            10:31:13  configure() uses liT now
 * 02-16-2001 15:29:25  reformat, stdtype.h
 * 02-20-2001 13:45:39  call LoadOpeningBook()
 * 02-21-2001 21:51:23  set and clear in_book, got_book flags
 * 03-09-2001 15:36:21  add ch_pending() function
 *            16:58:49  select() -> select_move() (conflict w/C)
 * 05-10-2001 16:17:03  remove old ch_pending fns
 * 05-11-2001  9:30:30  use SAN Kit I/O functions
 * 05-14-2001 11:03:06  add findSANMove()
 *            11:07:50  rearrange prototypes
 *            12:15:42  fix error in print_board() using puts()
 *            17:12:49  add san_flipit() for xboard setup
 * 05-15-2001  9:31:16  add SANcastle() for castle permissions
 * 05-16-2001  9:34:53  add debug print to findSANMove()
 * 05-17-2001 11:57:24  mask SAN Kit squares with sqM
 * 05-18-2001 11:32:08  call print_board() from JETFatal()
 * 05-23-2001 12:53:38  turn off output from think() in select_move()
 * 05-29-2001 14:53:13  clean up includes
 * 05-31-2001 10:17:14  sys/time.h -> time.h for MSC, others
 * 12-19-2001 14:07:00  restore ch_pending() function
 * 01-12-2001 10:20:00  use optnv[optn_verb] in call to think()
 * 02-06-2002 23:24:00  assert() bug in findSANMove()
 * 02-08-2002 20:02:00  add JETtoSAN(), change to JETFatal()
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>

#include "stdtype.h"

#ifndef _LINUX
#include "SanKit\sandef.h"
#include "SanKit\sanvar.h" 
#include "SanKit\sandsp.h"
#include "SanKit\sanutl.h"
#include "SanKit\sanmsc.h"
#else
#include "SanKit/sandef.h"
#include "SanKit/sanvar.h" 
#include "SanKit/sandsp.h"
#include "SanKit/sanutl.h"
#include "SanKit/sanmsc.h"
#endif


#include "jetdef.h"
#include "jetdat.h"
#include "jetpro.h"

/* Squares in the SAN kit are flipped from what the engine
uses. This array converts then. */
static siT san_flip[64] = 
  {
  56,  57,  58,  59,  60,  61,  62,  63,
  48,  49,  50,  51,  52,  53,  54,  55,
  40,  41,  42,  43,  44,  45,  46,  47,
  32,  33,  34,  35,  36,  37,  38,  39,
  24,  25,  26,  27,  28,  29,  30,  31,
  16,  17,  18,  19,  20,  21,  22,  23,
   8,   9,  10,  11,  12,  13,  14,  15,
   0,   1,   2,   3,   4,   5,   6,   7
  }
;

/*
 * Called when engine initialized first time
 */
void initialize(void)
  {
  init();
  gen();
  max_time = 30 * 10000; /* 30 seconds */
  max_depth = 4;
  got_book = LoadOpeningBook(DEFAULT_OPENING_BOOK);
  in_book = got_book;
  puts(got_book ? "Opening book loaded." : "Opening book NOT loaded.");
  }

/*
 * Called when SAN Kit exits
 */
void terminate(void)
  {
  }

/*
 * Called when new game begun
 */
void newg(void)
  {
  init();
  in_book = got_book;
  gen();
  }

/*
 * Called when move is played
 * Passed from and to squares as ints
 * (Requires promotion piece as int, too)
 * Note use of san_flip[], gen()
 */
void playmove(siT fromsq, siT tosq, siT promo)
  {
  siT i;
  boolT found;

  ply = 0;
  gen(); 
  found = FALSE;
  for (i = 0; i < first_move[1]; ++i)
    if (gen_dat[i].m.b.from == san_flip[fromsq] && 
      gen_dat[i].m.b.to == san_flip[tosq])
      {
      found = TRUE;
      /* get promotion piece right */
      if (gen_dat[i].m.b.bits & 32)
        i += (promo-1);
      break;
      }
  if (!found || !makemove(gen_dat[i].m.b))
    {
    SANDspLine("Illegal move");
    /* no gen() causes SAN kit to abort next move */
    }
  else
    {
    ply = 0;
    gen();
    }
  }

/*
 * Retract last move
 */
void unplaymove()
  {
  takeback();
  ply = 0;
  gen();
  }

/*
 * Choose best move from current position
 * Note use of san_flip[]
 * Now uses SAN Kit's "verb" option to control output information
 * This is needed by some testcases.
 */
void select_move(siT *from, siT *to, siT *promo)
  {
  move_bytes m;
  int post;

  post = (optnv[optn_verb]) ? 1 : 0;
  think(post);
  m = pv[0][0].b;
  *from = san_flip[m.from];
  *to = san_flip[m.to];
  *promo = (m.bits & 32) ? m.promote : 0;
  }

/*
 * Configure engine
 * Passed playing level (depth) and search time (milliseconds)
 */
void configure(siT depth, liT millisecs)
  {
  max_depth = depth;
  max_time = millisecs;
  }

/*
 * Setup board
 * Passed square, color, and piece.
 * Note square is a SAN kit square and must be flipped.
 * Color and piece are OK, though.
 */
void placeboard(siT sq, siT c, siT p)
  {
  color[san_flip[sq]] = c;
  piece[san_flip[sq]] = p;
  }

/*
 * Set color on move, and opposite color
 */
void placecolor(siT c)
  {
  side = c;
  xside = c ? LIGHT : DARK;
  }

/*
 * Set castle permissions. 
 * See file SANDAT.C for definitions of bits.
 */
void placecastle(siT m)
  {
  castle = m;
  }

/*
 * Set en passant target square.
 * Note that square passed is a SAN kit square, it gets flipped.
 */    
void placeep(siT sq)
  {
  ep = san_flip[sq];
  }

/*
 * Set halfmove clock (for draws) and fullmove number
 * Note we are no longer considered in book.
 */
void placeclocks(siT half, siT full)
  {
  fifty = half;
  hply = full;
  in_book = FALSE;
  }

/* get_ms() returns milliseconds elapsed since midnight,
January 1, 1970. */

#include <sys/timeb.h>
boolT ftime_ok = FALSE;  /* does ftime return milliseconds? */
liT get_ms()
  {
  struct timeb timebuffer;
  ftime(&timebuffer);
  if (timebuffer.millitm != 0)
    ftime_ok = TRUE;
  return (long)(timebuffer.time * 1000L) + timebuffer.millitm;
  }
	

/* print_board() prints board */
/* ?? should be default SAN printing function */

void print_board()
  {
  siT i;

	sprintf(stv, "\n8 ");
  SANDspSTV();
  for (i = 0; i < 64; ++i) 
    {
    switch (color[i]) 
      {
    case EMPTY:
      sprintf(stv, " .");
			SANDspSTV();
      break;
    case LIGHT:
      sprintf(stv, " %c", piece_char[piece[i]]);
			SANDspSTV();
      break;
    case DARK:
      sprintf(stv, " %c", piece_char[piece[i]] + ('a' - 'A'));
			SANDspSTV();
      break;
      }
    if ((i + 1) % 8 == 0 && i != 63)
			{
      sprintf(stv, "\n%d ", 7 - ROW(i));
			SANDspSTV();
			}
    }
	sprintf(stv, "\n\n   a b c d e f g h\n\n");
  SANDspSTV();
  }

/* move_str returns string with move m in coordinate notation */

charptrT move_str(move_bytes m)
  {
  static char str[6];

  char c;

  if (m.bits & 32) 
    {
    switch (m.promote) 
      {
    case KNIGHT:
      c = 'n';
      break;
    case BISHOP:
      c = 'b';
      break;
    case ROOK:
      c = 'r';
      break;
    default:
      c = 'q';
      break;
      }
    sprintf(str, "%c%d%c%d%c",
      COL(m.from) + 'a',
      8 - ROW(m.from),
      COL(m.to) + 'a',
      8 - ROW(m.to),
      c);
    }
  else
    sprintf(str, "%c%d%c%d",
      COL(m.from) + 'a',
      8 - ROW(m.from),
      COL(m.to) + 'a',
      8 - ROW(m.to));
  return str;
  }
	
/*
 * Flip squares between SAN Kit and Jester
 * Note this is inverse of normal function
 */
sqT san_flipit(sqT sq)
	{
	siT i;
	
	for (i = 0; i < 64; i++) 
		if (san_flip[i] == sq) 
			return i;
	SANFatal("san_flipit()");
	return -1;
	}
	
/*
 * Convert a move in JET format to one used by the SAN Kit.
 *  This should allow correct searching by SAN Kit functions.
 *  Converts squares, uses CURRENT color/piece values, and
 *  attempts to set m_scmv from bits.
 */
void JETtoSAN(move_bytes b, mptrT mptr)
  {
  mptr->m_flag = 0; 
  mptr->m_frsq = san_flipit(b.from);
  mptr->m_tosq = san_flipit(b.to); 
  mptr->m_frcp = (color[b.from] == EMPTY) ? cp_v0 : (piece[b.from] + 6*color[b.from]); 
  mptr->m_tocp = (color[b.to] == EMPTY) ? cp_v0 : (piece[b.to] + 6*color[b.to]); 
  if (b.bits & 2)
    mptr->m_scmv = (b.to == G8 || b.to == G1) ? scmv_cks : scmv_cqs; 
  else if (b.bits & 4)
    mptr->m_scmv = scmv_epc;
  else if (b.bits & 32)
    mptr->m_scmv = b.promote + 3;
  else
    mptr->m_scmv = 0;
  }
	
/*
 * Convert Jester castle mask into SAN Kit castle mask
 */
siT san_castle(siT c)
	{
	siT sc;
	
	sc = 0;
	sc |= (c & 1) ? cast_wk : 0;
	sc |= (c & 2) ? cast_wq : 0;
	sc |= (c & 4) ? cast_bk : 0;
	sc |= (c & 8) ? cast_bq : 0;
	return sc;
	}
	
/*
 * JETFatal(): fatal error handler for engine
 * Allows for termination of xboard, etc.
 */
void JETFatal(void)
	{
	print_board();
	SANDspStatus();
	if (in_xboard) 
		xboard_resign();
	}

boolT ch_pending(void)
  {
  int is_ready(int);

  if (in_xboard)
    return is_ready(0); /* check STDIN for character pending in xboard */
  else
    return 0;
  }
	
/*<<< jester.c: EOF */
