/*>>> jetxbd.c: Jester chess engine Xboard interface */

/*
 *	JETXBD.C
 *  Jester Xboard interface
 *  Activate from command line with "xbrd" command
 *
 *  02-14-2001  9:13:48 Create file
 *             15:36:39 max_time liT from 'time' command
 *  02-15-2001 10:36:37 support 'level' command and time controls
 *             13:56:23 fix time calculation past first time control
 *             14:06:16 don't use 1 << 25 for liT time (?)
 *  02-16-2001 14:55:07 reformat, stdtype.h
 *  02-17-2001 11:10:00 fix time control calculation
 *  02-20-2001 20:27:24 modify time control calculation
 *  02-21-2001  8:58:24 fix "black" command for opening book
 *  02-21-2001 12:58:55 adjust time control by TIME_MARGIN
 *  02-21-2001 21:52:33 set in_book from got_book on "new"
 *  03-05-2001 15:39:57 Edit and Analyze modes supported
 *  03-07-2001  9:54:38 "post" is now static
 *  03-13-2001  9:54:46 force search time in analysis mode to x seconds
 *             11:59:04 fix error reading '.' from xboard in analysis mode
 *  05-09-2001 10:49:07 max_time=3 sec. in analysis mode (kludge)
 *  05-11-2001 10:08:58 use SAN Kit I/O functions
 *  05-11-2001 14:12:15 remove fflush(stdout)
 *  05-14-2001  9:43:29 max_time=5 sec. in analysis mode (kludge)
 *             11:07:28 rearrange prototypes
 *             14:07:13 make moves in SAN Kit, too
 *             14:19:26 refuse and allow JET calls as needed
 *             15:41:42 call SAN Kit fns when setting up board
 *  05-15-2001  9:06:55 fix error in edit() mode
 *  05-18-2001 11:35:47 add xboard_resign()
 *  05-21-2001 10:55:32 update SAN board fixed
 *  05-24-2001 13:20:00 support forced move cache
 *  12-19-2001 13:58:00 restore "real" analysis mode time settings
 *  01-09-2001 20:37:00 support time increment in "level" command
 *  02-06-2002 12:25:00 fix set_time_control() via Dann Corbit 
 *  02-06-2002 19:01:00 call SANInitGround() "new"
 *  02-06-2002 23:43:00 ignore "result" command
 * 	02-08-2002 18:44:00 use JETtoSAN() for SAN move handling
 * 	           21:35:00 don't call JETtoSAN() for bad commands
 *             22:17:00 set lower limit for max_time, 1 sec.
 *  02-10-2002 10:58:00 toggle SAN active color in edit mode
 */

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "stdtype.h"

#ifndef _LINUX
#include "SanKit\sandef.h"
#include "SanKit\sanvar.h"
#include "SanKit\sandsp.h"
#include "SanKit\sanosi.h"
#include "SanKit\sanmpu.h"
#include "SanKit\sanval.h"
#include "SanKit\sanint.h"
#include "SanKit\sanhst.h"
#else
#include "SanKit/sandef.h"
#include "SanKit/sanvar.h"
#include "SanKit/sandsp.h"
#include "SanKit/sanosi.h"
#include "SanKit/sanmpu.h"
#include "SanKit/sanval.h"
#include "SanKit/sanint.h"
#include "SanKit/sanhst.h"
#endif

#include "jetdef.h"
#include "jetdat.h"
#include "jetpro.h"
#include "jester.h"

#define TIME_MARGIN 1000 /* allow this time in ms for delays, etc. */
#define ETERNITY (1 << 25)
#define BOTTOMLESS 32
#define SHALLOW 4

static siT post = 0;
static siT computer_side;
static siT control_moves;

/* pass position to SAN Kit */

static sqT sq;
static bT new_b;
static cT new_actc;
static castT new_cast;
static sqT new_epsq;
static siT new_hmvc, new_fmvn;

/*
 * move_number() returns the fullmove number using hply.  It is
 * numbered from 1.
 */
siT move_number()
  {
  return (hply/2)+1;
  }

/*
 * Set time control from "level" command.
 * Sets max_depth, max_time, and control_moves.
 * Used in both normal and analysis modes.
 * Fixed error in division 2-6-2002: thanks, Dann Corbit!
 */	
void set_time_control(char *line)
	{
	int first, second, third;
  
	sscanf(line, "level %d %d %d", &first, &second, &third);
	control_moves = third ? 30 : first ;
	control_moves = control_moves ? control_moves : 60; 
	max_depth = BOTTOMLESS;
  max_time = max(TIME_MARGIN, (second * 60)/control_moves * 1000); /* ms. */ 
	}

/*
 * print_result() checks to see if the game is over, and if so,
 * prints the result.
 */
void print_result()
  {
  siT i;

  /* is there a legal move? */
  for (i = 0; i < first_move[1]; ++i)
    if (makemove(gen_dat[i].m.b))
      {
      takeback();
      break;
      }
  if (i == first_move[1])
    {
    if (in_check(side))
      {
      if (side == LIGHT)
        SANDspStr("0-1   {Black mates  }\n");
      else
        SANDspStr("1-0   {White mates  }\n");
      }
    else
      SANDspStr("1/2-1/2   {Stalemate  }\n");
    }
  else if (reps() == 3)
    SANDspStr("1/2-1/2   {Draw by repetition  }\n");
  else if (fifty >= 100)
    SANDspStr("1/2-1/2   {Draw by fifty move rule  }\n");
  }

/*
 * Return true if piece p of color c is at square sq.
 * s.b. a macro
 */	
boolT isPieceAt(siT sq, siT p, siT c)
	{
	return ((color[sq] == c) && (piece[sq] == p)) ? TRUE : FALSE;
	}
	
/*
 * Undo a single move
 * Used in both "normal" and "analysis" modes.
 */
void undo_move(void)
	{
	if (hply)
		{
		takeback();
		SANUnplayMove();
		ply = 0;
		gen();
		}
	}

/*
 * Make a move in Xboard mode.
 * Passed line containing move in coordinate notation
 * Returns true if move made OK, false otherwise (end of game)
 */	
boolT make_xmove(char line[])
	{
	mT SANmove;
	boolT found;
	siT i, from, to;
	
	from = line[0] - 'a';
	from += 8 * (8 - (line[1] - '0'));
	to = line[2] - 'a';
	to += 8 * (8 - (line[3] - '0'));
	found = FALSE;
	for (i = 0; i < first_move[1]; ++i)
		if (gen_dat[i].m.b.from == from && gen_dat[i].m.b.to == to)
			{
			found = TRUE;
			if (gen_dat[i].m.b.bits & 32)
				switch (line[4])
					{
				case 'n':
					break;
				case 'b':
					i += 1;
					break;
				case 'r':
					i += 2;
					break;
				default:
					i += 3;
					break;
					}
			break;
			}
	if (found)
		{
		JETtoSAN(gen_dat[i].m.b, &SANmove);
		}
	if (!found || !makemove(gen_dat[i].m.b))
		{
		sprintf(stv, "Error (unknown command): %s\n", line);
		SANDspSTV();
		return FALSE;
		}
	else
		{
		if (forced.u != gen_dat[i].m.u)
			forced_reply.u = 0L;
		SANPlayMove(&SANmove);
		ply = 0;
		gen();
		print_result();
		return TRUE;
		}
	}

/*
 * Fix castling permissions once a position has been set up.
 * Looks to see if Rooks and King have moved from original squares.
 * Not foolproof, but works for most positions.
 */
siT fix_castle(void)
	{
	siT new_mask;
	
	new_mask = 0;
	if (isPieceAt(E8, KING, DARK))
		{
		if (isPieceAt(A8, ROOK, DARK))
			new_mask += 8;
		if (isPieceAt(H8, ROOK, DARK))
			new_mask += 4;
		}
	if (isPieceAt(E1, KING, LIGHT))
		{
		if (isPieceAt(A1, ROOK, LIGHT))
			new_mask += 2;
		if (isPieceAt(H1, ROOK, LIGHT))
			new_mask += 1;
		}
	return new_mask;
	}
	
/* 
 * Update SAN board
 * Copies values set to SAN Kit board values, then calls Kit
 */
void updateSAN(void)
	{	
  for (sq = sq_a1; sq <= sq_h8; sq++)
    gb.bv[sq] = new_b.bv[sq];
  curr_e.e_actc = new_actc;
  curr_e.e_pasc = inv_cv[new_actc];
  curr_e.e_cast = new_cast;
  curr_e.e_epsq = new_epsq;
  curr_e.e_hmvc = new_hmvc;
  curr_e.e_fmvn = new_fmvn;
  SANZapUp();
	}
	
/*
 * Handle Xboard Edit Mode.
 * This allows positions to be set up and analyzed.
 */
void edit(void)
	{
	siT ccolor;
  char line[256], command[256];
	
  SANDspNL();
	ccolor = LIGHT;
	in_book = FALSE;
	for (;;)
		{
    SANfgets(line, 256);
    if (line[0] == '\n')
      continue;
    sscanf(line, "%s", command);
		if (!strcmp(command, "."))
			{
			/* fix castle permissions */
			castle = fix_castle();
			new_cast = san_castle(castle);
			
		 	/* exit edit mode */
			ply = 0;
			hply = 0;
			gen();
			
			/* update SAN Kit board */
			updateSAN();
			return;
			}
	 	if (!strcmp(command, "#"))
			{
			/* reset board(s) */
			siT i;
			for (i = 0; i < 64; i++)
				{
		    color[i] = piece[i] = EMPTY;
				}
			ep = -1;
			fifty = 0;
			hply = 0;

			/* update SAN Kit values */
		  for (sq = sq_a1; sq <= sq_h8; sq++)
    		new_b.bv[sq] = cp_v0;
			new_cast = 0;
			new_epsq = -1;
			new_fmvn = 2;
			new_hmvc = 0;
			new_actc = LIGHT; /* assumed */
			continue;
			}
		if (!strcmp(command, "c"))
			{
			/* toggle color */
			ccolor = (ccolor == LIGHT) ? DARK : LIGHT;
			new_actc = ccolor;
			continue;
			}

		/* place piece on square */
		if (strlen(command) == 3)
			{
			siT what = EMPTY;
			siT j, p, where; 
			p = command[0];
			for (j = 0; j < 6; j++)
				{
				if (p == piece_char[j]) 
					{
					what = j;
					break;
					}
				}
				
			/* square is 2nd and 3rd chars */
	    where = command[1] - 'a';
	    where += 8 * (8 - (command[2] - '0'));
			
			/* update board */
			piece[where] = what;
			color[where] = ccolor;
			
			/* update SAN board */
			new_b.bv[san_flipit(where)] = (ccolor == LIGHT) ? what : what+6;
			continue;
			}
		}
	}
	
/*
 * Handle XBoard analysis mode
 */
void analyze(void)
	{
	liT save_time;
	siT save_post, save_depth;
	boolT save_in_book, result;
  char line[256], command[256];
	
  SANDspNL();
	
	/* save prior engine settings */
	save_in_book = in_book;
	save_post = post;
	save_time = max_time;
	save_depth = max_depth;
	
	/* change settings for analysis mode */
	max_time = ETERNITY;  /* was ETERNITY */
	max_depth = BOTTOMLESS;
	in_book = FALSE;
	post = 2;
	result = FALSE;	
	ply = 0;
	gen();
	
	/* do analysis until told to stop */
	while (!result)
		{
		think(post);
		if (!pv[0][0].u)
			{
			if (!result)
				{
				print_result();
				result = TRUE;
				}
			}
		ply = 0;
		gen();
		
		/* process commands in analysis mode */
read_again:		
    SANfgets(line, 256);
    if (line[0] == '\n' || line[0] == '.')
			goto read_again;
    sscanf(line, "%s", command);
		if (!strcmp(command, "exit"))
			{
			max_depth = save_depth;
			max_time = save_time;
			post = save_post;
			in_book = save_in_book;
			return;
			}
		if (!strcmp(command, "new"))
			{
			SANInitGround();
			init();
			max_time = ETERNITY; /* was: ETERNITY */
			max_depth = BOTTOMLESS;
			in_book = FALSE;
			post = 2;
			ply = 0;
			gen();
			result = FALSE;
			continue;
			}
		if (!strcmp(command, "undo"))
			{
			undo_move();
			continue;
			}
		if (!strcmp(command, "edit"))
			{
			edit();
			continue;
			}
		if (!strcmp(command, "easy"))
			{
			max_time = 1000; /* 1 second */
			max_depth = SHALLOW; 
			continue;
			}
		if (!strcmp(command, "hard"))
			{
			max_time = ETERNITY; /* was: ETERNITY */
			max_depth = BOTTOMLESS;			
			continue;
			}
		if (!strcmp(command, "level"))
			{
			set_time_control(line);
			continue;
			}
		/* else process move and print result, if any */
		make_xmove(line);
		}
	}
	
/* 
 * xboard_resign() abruptly resigns game WHEN A FATAL ERROR OCCURS
 */
void xboard_resign(void)
	{
	sprintf(stv, "%s {%s resigns}\n", 
	  ((computer_side == LIGHT) ? "0-1" : "1-0"), 
	  ((computer_side == LIGHT) ? "White" : "Black"));
	SANDspSTV();
	fflush(stdout);
  }

/*
 * xboard() is a substitute for main() that is XBoard and
 * WinBoard compatible. See the following page for details:
 * www.research.digital.com/SRC/personal/mann/xboard/engine-intf.html
 */
void xboard()
  {
	siT moves;
	mT SANmove;
  siT post = 0;
  int time, inc;
  bool icc = false;
  char line[256], command[256], msg[356];

	SANXboardInit();
	JETRefuseCalls();
  SANDspNL();
  init();
  gen();
  computer_side = EMPTY;
  for (;;)
    {
    if (side == computer_side)
      {
			if (!in_book && forced_reply.u) 
				{
				pv[0][0].u = forced_reply.u;
				forced_reply.u = 0L;
  	    }
			else 
	      think(post);
      if (!pv[0][0].u)
        {
        computer_side = EMPTY;
        continue;
        }
			JETtoSAN(pv[0][0].b, &SANmove);
      makemove(pv[0][0].b);
			SANPlayMove(&SANmove);
      sprintf(stv, "move %s\n", move_str(pv[0][0].b));
			SANDspSTV();
      ply = 0;
      gen();
      print_result();
      continue;
      }
    SANfgets(line, 256);
    if (line[0] == '\n')
      continue;
    sscanf(line, "%s", command);
   
    if (!strcmp(command, "xboard")) {
      printf("done=0\n");
      continue;
      }
      
    if (!strcmp(command, "new"))
      {
			SANInitGround();
      init();
      gen();
      computer_side = DARK;
		  in_book = got_book;
      continue;
      }
      
      if (!strcmp(command, "protover")) {
            printf("\nfeature colors=1\n");
            printf("feature myname=\"Jester %s by %s\"\n",JESTER_VERSION, AUTHOR);
            printf("feature sigint=0\n");
            printf("feature sigterm=0\n");
            printf("feature time=1\n");
            printf("feature pause=0\n");
            printf("feature variants=\"normal,nocastle\"\n");
            printf("feature ping=0\n");
            printf("feature random=0\n");
            printf("feature name=0\n"); 
            printf("feature reuse=0\n"); 
            printf("feature analyze=1\n");
            printf("feature ics=1\n");
            printf("feature done=1\n");
            continue;
}
      
      
    if (!strcmp(command, "quit"))
			{
			SANXboardTerm();
			JETAllowCalls();
      return;
			}
    if (!strcmp(command, "force"))
      {
      computer_side = EMPTY;
      continue;
      }
    if (!strcmp(command, "white"))
      {
      side = LIGHT;
      xside = DARK;
      ply = 0;
      hply = 0;
      gen();
      computer_side = DARK;
			SANPutACTC(c_w);
      continue;
      }
    if (!strcmp(command, "black"))
      {
      side = DARK;
      xside = LIGHT;
      ply = 0;
      hply = 1;
      gen();
      computer_side = LIGHT;
			SANPutACTC(c_b);
      continue;
      }
    if (!strcmp(command, "st"))
      {
      sscanf(line, "st %ld", &max_time);
      max_time *= 1000;
      max_depth = BOTTOMLESS;
      continue;
      }
    if (!strcmp(command, "sd"))
      {
      sscanf(line, "sd %d", &max_depth);
      max_time = ETERNITY; /* was: 1 << 25 */
      continue;
      }
    if (!strcmp(command, "level"))
      {
		sscanf(line, "level %d %d %d", &time, &time, &inc);
      
	control_moves = inc ? 30 : time ;
	control_moves = control_moves ? control_moves : 60; 
	max_depth = BOTTOMLESS;
    max_time = max(TIME_MARGIN, (time * 60)/control_moves * 1000); /* ms. */ 
    continue;
      }
      
      	if (!strcmp (command, "time")) {
        sscanf(line, "time %d", &time); 
	    max_time = (CLOCKS_PER_SEC/100) * (time/16) + (inc * CLOCKS_PER_SEC);
        if (move_number() != control_moves)
        moves = control_moves - (move_number() % control_moves);
        else
        moves = control_moves;
        max_depth = BOTTOMLESS;     
        continue;
       }
       
       
       if (!strcmp(command, "told")) {
			// told via "tell" msg or left a msg via "message"
			strcpy(msg, (line+5));
			printf("tellics kibitz %s\n",msg);
			printf("tellics draw\n");
			continue;
			}

		if (!strcmp(command, "whispered")) {
		// whispered to via "whisper"
		continue;
		}

		if (!strcmp(command, "heardkibitz")) {
			// picked up a kibitz
		continue;
		}
		
		if (!strcmp(command, "heardshout")) {
		continue;
		}
		
		if (!strcmp(command, "heardsshout")) {
		continue;
		}

          if (!strcmp(command, "ics")) { // are we on an internet chess server?
			if(line[4] == '-') {
				icc = false;
			} else {
				icc = true;
			}
			continue;
		}
	
      
      
    if (!strcmp(command, "go"))
      {
      computer_side = side;
      continue;
      }
    if (!strcmp(command, "hint"))
      {
      think(0);
      if (!pv[0][0].u)
        continue;
      sprintf(stv, "Hint: %s\n", move_str(pv[0][0].b));
			SANDspSTV();
      continue;
      }
    if (!strcmp(command, "undo"))
      {
			undo_move();
      continue;
      }
    if (!strcmp(command, "remove"))
      {
      if (hply < 2)
        continue;
      takeback();
			SANUnplayMove();
      takeback();
			SANUnplayMove();
      ply = 0;
      gen();
      continue;
      }
    if (!strcmp(command, "post"))
      {
      post = 2;
      continue;
      }
    if (!strcmp(command, "nopost"))
      {
      post = 0;
      continue;
      }
		if (!strcmp(command, "edit"))
			{
			edit();
			continue;
			}
		if (!strcmp(command, "analyze"))
			{
			analyze();
			continue;
			}
			
			
	 if (!strcmp(command, "name")) { 
     char name[265];
     sscanf(line,"name %s",name);
     printf("opponent's name is %s\n",name);
     if (icc) printf("tellics say Jester is playing %s\n",name);
	 continue;
}
		
	
     if (!strcmp(command,"variant")) {      
     char variant[256];
     sscanf(line, "variant %s",variant);
     if (strcmp(variant,"normal") || strcmp(variant,"nocastle")) {
     printf("unknown variant\n");
     }
     continue;
     }		
	
	if (!strcmp(command, "result")) {
        char result[256];
        sscanf(line,"result %s",result);
        computer_side = EMPTY;
        continue;
        }
        
        
     if (!strcmp(command, "accepted") || !strcmp(command, "rejected")
         ||  !strcmp(command, "computer") || !strcmp(command, "hard")
         ||  !strcmp(command, ".")        || !strcmp(command, "easy")
         ||  !strcmp(command, "?")        || !strcmp(command, "random")
         ||  !strcmp(command, "otim")) {
            
             continue;
         }
		    
	
	
		make_xmove(line);
    }
  }

/* jetxbd.c: EOF */
